/*
 * Licensed Materials - Use restricted, please refer to the "Samples Gallery" terms and conditions
 * in the IBM International Program License Agreement.  Copyright IBM Corporation 2003, 2005. All
 * Rights Reserved.
 */
package com.ibm.jps.architect.proc;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Component;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Interface;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Namespace;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.ParameterDirectionKind;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Relationship;
import org.eclipse.uml2.uml.Type;
import org.eclipse.uml2.uml.VisibilityKind;

import com.ibm.jps.architect.util.RSMModel;
import com.ibm.jps.architect.util.RSMRelation;
import com.ibm.jps.architect.util.RSMUtil;

public class AccessViolations {
	private String curNSName;
	private String curNSType;
	private boolean fixAccessViolations = false;
	private boolean hasChanged;
	private boolean headerPrinted;
	private boolean isDebug = true;
	private boolean isExternalLibrary;
	private int numViolations;
	private int totViolations;
	private int numFixedViolations;
	private int totFixedViolations;
	private int numPackages;
	private List<String> newDependencies = new ArrayList<String>();
	private List<String> illegalDependencies = new ArrayList<String>();
	private List<String> authorizedDependencies = new ArrayList<String>();
	private List<String> prohibitedDependencies = new ArrayList<String>();
	private List<Package> globalLibraries = new ArrayList<Package>();
	private List<Package> globalLayers = new ArrayList<Package>();
	private MessageConsoleStream console = null;
	private Shell shell = null;
	private IResource mRsrc = null;

	public AccessViolations(MessageConsoleStream inConsole) {
		super();
		this.shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
		this.console = inConsole;
	}

	public boolean execute(Namespace inNS) throws CancelException {
		// Marker management
		mRsrc = RSMModel.getModelResource(inNS);
		if (mRsrc != null) {
			try {
				mRsrc.deleteMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
			} catch (CoreException e) {
				log("??? Error deleting markers from Problems view.");
			}
		}
		
		totViolations = 0;
		totFixedViolations = 0;
		numPackages = 0;
		hasChanged = false;
		initLists(inNS.getModel());
		processNS(inNS, false);
		return hasChanged;
	}

	/**
	 * 
	 */
	private void initLists(Model inModel) {
		if (!newDependencies.isEmpty()) newDependencies.clear();
		if (!illegalDependencies.isEmpty()) illegalDependencies.clear();
		if (!authorizedDependencies.isEmpty()) authorizedDependencies.clear();
		if (!prohibitedDependencies.isEmpty()) prohibitedDependencies.clear();
		// Create the list of global lists (must be done before calling
		// initDependencyLists)
		for (Object curPkg_o:inModel.getNestedPackages()) {
			populateGlobalLists((Package) curPkg_o);
		}
		for (Object curPkg_o:inModel.getNestedPackages()) {
			initDependencyLists((Package) curPkg_o);
		}
	}

	/**
	 * @param package1
	 */
	private void populateGlobalLists(Package inPkg) {
		if (inPkg.getAppliedStereotype("ArchitectAssistant::global") != null) {
			if (inPkg.getAppliedStereotype("ArchitectAssistant::layer") != null) {
				globalLayers.add(inPkg);
				if (isDebug) System.out.println(">>> populateGlobalLists: " + inPkg.getName() + " added to globalLayers");
			} else {
				globalLibraries.add(inPkg);
				if (isDebug) System.out.println(">>> populateGlobalLists: " + inPkg.getName() + " added to globalLists");
			}
		}
		for (Object curPkg_o:inPkg.getNestedPackages()) {
			populateGlobalLists((Package) curPkg_o);
		}
	}

	/**
	 * @param inPkg
	 */
	private void initDependencyLists(Package inPkg) {
		if (inPkg.getAppliedStereotype("ArchitectAssistant::layer") != null) {
			String pkgName = RSMModel.getNameWithoutModel(inPkg.getQualifiedName());
			for (Package p:globalLibraries) {
				authorizedDependencies.add(pkgName + ">" + RSMModel.getNameWithoutModel(p.getQualifiedName()));
			}
			List<RSMRelation> deps = RSMModel.getDependencies(inPkg);
			for (RSMRelation rel:deps) {
				NamedElement supplier = rel.getSupplier();
				if (supplier instanceof Package) {
					Package supplierPkg = (Package) supplier;
					authorizedDependencies.add(pkgName + ">"
							+ RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName()));
					prohibitedDependencies.add(RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName()) + ">"
							+ pkgName);
				}
			}
			for (Package p:globalLayers) {
				if (inPkg != p) {
					authorizedDependencies.add(pkgName + ">" + RSMModel.getNameWithoutModel(p.getQualifiedName()));
					prohibitedDependencies.add(RSMModel.getNameWithoutModel(p.getQualifiedName()) + ">" + pkgName);
				}
			}
		}
		for (Object curPkg_o:inPkg.getNestedPackages()) {
			initDependencyLists((Package) curPkg_o);
		}
	}

	/**
	 * @param inStr
	 */
	private void log(String inStr) {
		if (inStr.startsWith("!!!")) {
			RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_WARNING);
		} else if (inStr.startsWith("***")) {
			RSMModel.addMarker(mRsrc, inStr.substring(4), IMarker.SEVERITY_ERROR);
		}
		console.println(inStr);
		if (isDebug) System.out.println(inStr);
	}

	/**
	 * @param inNS
	 * @throws CancelException
	 * @throws BrokenDependencyException
	 */
	private void processNS(Namespace inNS, boolean inParentIsExternal) throws CancelException {
		boolean curIsExternal = ((inNS.getAppliedStereotype("ArchitectAssistant::externalLibrary") != null) ||
				(inNS.getAppliedStereotype("Standard::modelLibrary") != null) );
		if (curIsExternal) {
			isExternalLibrary = true;
		} else if (!inParentIsExternal) 
			isExternalLibrary = false;
		curNSName = inNS.getQualifiedName();
		curNSType = inNS.eClass().getName().toLowerCase();
		if (isDebug) System.out.println(">>> Processing " + RSMModel.getNameWithoutModel(curNSName) + "; isExternalLibrary = " + isExternalLibrary);
		headerPrinted = false;
		numViolations = 0;
		numFixedViolations = 0;
		numPackages++;
		processClassifiers(inNS);
		totViolations = totViolations + numViolations;
		totFixedViolations = totFixedViolations + numFixedViolations;
		if (headerPrinted) {
			if (fixAccessViolations)
				log("    " + numFixedViolations + " access violation(s) resolved, " + numViolations + " not resolved");
			else
				log("    " + numViolations + " access violation(s) found");
		}
		// Process nested packages
		if (inNS instanceof Package) {
			Package p = (Package) inNS;
			for (Object obj:p.getNestedPackages()) {
				processNS((Package) obj, inParentIsExternal || curIsExternal);
			}
			// Build a list of components owned by the package: this is necessary because we get to the components through
			// getOwnedElements(). But since this list can be modified as a result of adding dependencies, it can result in
			// an exception
			List<Component> components = new ArrayList<Component>();
			for (Object obj : p.getOwnedElements()) {
				if (obj instanceof Component) {
					components.add((Component) obj);
				}
			}
			for (Component obj : components) {
				processNS(obj, inParentIsExternal || curIsExternal);
			}
		}
	}

	/**
	 * For each classifier of the specified package, check all the relationships of
	 * this classifier with other classifiers, as well as the dependencies of the
	 * associated component with other classes. For each relationship, determine
	 * if there is a dependency between the corresponding packages. If not, the
	 * "access violation" should be processed based on the context and the Fix
	 * Access Violations option.
	 * 
	 * @param inNS
	 * @throws BrokenDependencyException
	 * @throws CancelException
	 */
	private void processClassifiers(Namespace inNS) throws CancelException {
		// Build a list of components owned by the package: this is necessary because we get to the components through
		// getOwnedElements(). But since this list can be modified as a result of adding dependencies, it can result in
		// an exception
		List<Classifier> classifiers = new ArrayList<Classifier>();
		for (Object obj : inNS.getOwnedElements()) {
			if (obj instanceof Class || obj instanceof Component || obj instanceof Interface) {
				classifiers.add((Classifier) obj);
			}
		}
		for (Classifier obj : classifiers) {
			processClassifier(obj, inNS);
		}
	}

	/**
	 * processClassifier browses through all the relationships of the specified class.
	 * 
	 * @param c
	 * @throws CancelException
	 */
	private void processClassifier(Classifier inClient, Namespace inNS) throws CancelException {
		if (isDebug) System.out.println(">>> Processing " + inClient.eClass().getName().toLowerCase() + " "
			+ inClient.getQualifiedName());

		// Processing all relationships
		List<RSMRelation> allRels = RSMModel.getRelations(inClient);
		for (RSMRelation rel:allRels) {
			NamedElement supplier = rel.getSupplier();
			if (supplier == null) {
				processBrokenDependency(rel, inClient);
			} else {
				Package supplierPkg = supplier.getNearestPackage();
				if (isDebug)System.out.println(">>> processClassifier: " + rel.getRel().eClass().getName().toLowerCase() + " from "
					+ inClient.getName() + " (in " + RSMModel.getNameWithoutModel(inNS.getQualifiedName())
					+ ") to " + supplier.getName() + " (in " + RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName()) + ")");
				if (!dependencyExists(inNS, supplierPkg))
					processViolation(rel, inClient, inNS, supplier, supplierPkg);
			}
		}

		// Process attributes
		for (Object curProp_o:inClient.getAttributes()) {
			Property curProp = (Property) curProp_o;
			if (curProp.getAssociation() == null) {
				Type attType = curProp.getType();
				if (attType != null) {
					Package attPkg = attType.getPackage();
					if (isValidPackage(attPkg)) {
						if (isDebug) System.out.println(">>> processClassifier: attribute " + curProp.getName() + ":"
							+ attType.getName() + " (in " + RSMModel.getNameWithoutModel(attPkg.getQualifiedName()) + ")");
						if (!dependencyExists(inNS, attPkg))
							processViolation(curProp, inClient, inNS, curProp, attPkg);
					}
				}
			}
		}

		// Process all operations
		if (inClient instanceof Class || inClient instanceof Component || inClient instanceof Interface) {
			List operations;
			if (inClient instanceof Class)
				operations = ((Class) inClient).getOwnedOperations();
			else
				operations = ((Interface) inClient).getOwnedOperations();
			for (Object curOp_o:operations) {
				Operation curOp = (Operation) curOp_o;
				for (Object curParam_o:curOp.getOwnedParameters()) {
					Parameter curParam = (Parameter) curParam_o;
					Type paramType = curParam.getType();
					if (paramType != null) {
						Package paramPkg = paramType.getPackage();
						if (isValidPackage(paramPkg)) {
							if (isDebug)System.out.println(">>> processClassifier: operation " + curOp.getName() + ", parameter "
								+ curParam.getName() + ":" + paramType.getName() + " (in " + RSMModel.getNameWithoutModel(paramPkg.getQualifiedName()) + ")");
							if (!dependencyExists(inNS, paramPkg))
								processViolation(curOp, inClient, inNS, curParam, paramPkg);
						}
					}
				}
			}
		}
	}

	/**
	 * @param attPkg
	 * @return
	 */
	private boolean isValidPackage(Package inPkg) {
		if (inPkg != null) {
			String pkgName = inPkg.getQualifiedName();
			if (!pkgName.equals("UMLPrimitiveTypes") && !pkgName.equals("Java")
					&& !(pkgName.startsWith("Rose") && pkgName.endsWith("Datatypes"))
					&& !pkgName.equals("XSDDataTypes")) return true;
		}
		return false;
	}

	/**
	 * 
	 */
	private void processBrokenDependency(RSMRelation inRel, NamedElement inClient) throws CancelException {
		Relationship r = inRel.getRel();
		String s = r.eClass().getName().toLowerCase() + " from " + inClient.eClass().getName().toLowerCase() + " '"
				+ inClient.getName() + "' (in '"
				+ RSMModel.getNameWithoutModel(inClient.getNearestPackage().getQualifiedName()) + "')";
		if (fixAccessViolations) {
			MessageBox box = new MessageBox(shell, SWT.YES | SWT.NO | SWT.CANCEL | SWT.APPLICATION_MODAL | SWT.ICON_QUESTION);
			box.setText("AccessViolations Module");
			box.setMessage("There is a broken " + s + ".\n\nDo you want " + "to delete it?");
			int answer = box.open();
			if (answer == SWT.YES) {
				r.destroy();
				hasChanged = true;
				log("--- Broken " + s + ". Deleted at user's request.");
			} else if (answer == SWT.CANCEL) {
				throw new CancelException();
			} else
				log("!!! Broken " + s);
		}
	}

	private boolean dependencyExists(Namespace inFrom, Namespace inTo) {
		// For external libraries, only catch dependencies on packages that are
		// not external libraries themselves
		if (isExternalLibrary && inTo instanceof Package) {
			if (RSMModel.isExternalLibrary((Package) inTo)) return true;
		}
		if (!newDependencies.isEmpty()) {
			String depName = RSMModel.getNameWithoutModel(inFrom.getQualifiedName()) + ">"
					+ RSMModel.getNameWithoutModel(inTo.getQualifiedName());
			if (newDependencies.contains(depName)) return true;
		}
		return RSMModel.dependencyExists(inFrom, inTo);
	}

	private void processViolation(Object inElt, NamedElement inClient, Namespace inClientNS, NamedElement inSupplier,
			Namespace inSupplierNS) throws CancelException {
		String clientNSName = RSMModel.getNameWithoutModel(inClientNS.getQualifiedName());
		String supplierNSName = RSMModel.getNameWithoutModel(inSupplierNS.getQualifiedName());
		// if (dependencyAuthorized(clientNSName, supplierNSName)) return;
		if (isDebug) System.out.println(">>> processViolation: " + clientNSName + ">" + supplierNSName);

		// Log a violation if:
		// 1) This is a prohibited dependency, or
		// 2) The Fix Access Violations option is not selected, or
		// 3) The client is contained within the supplier (relationship from a
		// package to one of
		// its parent packages), or
		// 4) The relation is already logged as an illegal relationship (as a
		// result of the user
		// not accepting to create a Dependency between the packages from a
		// previous relation).
		String msg = depInfo(inElt, inClient, inSupplier);
		if (dependencyProhibited(clientNSName, supplierNSName)) {
			addViolation(msg + ", reason=prohibited dependency");
		} else if (RSMModel.containedIn(clientNSName, supplierNSName)) {
			addViolation(msg + ", reason=the supplier is the parent of the client");
		} else if (invalidVisibility(inClientNS, inSupplierNS)) {
			addViolation(msg + ", reason=supplier is not visible to client");
		} else if (!fixAccessViolations) {
			addViolation(msg);
		} else if (illegalDependencies.contains(clientNSName + ">" + supplierNSName)) {
			addViolation(msg + ", reason=previously disallowed by user");
		} else {
			fixViolation(inClientNS, inSupplierNS, msg);
		}
	}

	/**
	 * @param clientPkg
	 * @param supplierPkg
	 * @return
	 */
	private boolean invalidVisibility(Namespace clientPkg, Namespace supplierPkg) {
		if (supplierPkg.getVisibility() == VisibilityKind.PUBLIC_LITERAL)
			return false;
		else if (supplierPkg.getVisibility() == VisibilityKind.PRIVATE_LITERAL) {
			// Client and supplier must belong to the same parent package
			Package clientParent = clientPkg.getNearestPackage();
			Package supplierParent = supplierPkg.getNearestPackage();
			if (clientParent == null && supplierParent == null) {
				return false;
			} else if (clientParent != null && supplierParent != null) {
				if (clientParent == supplierParent) return false;
			}
		} else {
			// TODO Handle protected and package visibility
		}
		return true;
	}

	private void fixViolation(Namespace clientPkg, Namespace supplierPkg, String msg) throws CancelException {
		String clientPkgName = RSMModel.getNameWithoutModel(clientPkg.getQualifiedName());
		String supplierPkgName = RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName());
		if (isDebug) System.out.println(">>> fixViolation: " + clientPkgName + " to " + supplierPkgName);
		if (dependencyAuthorized(clientPkgName, supplierPkgName))
			addDependency(clientPkg, supplierPkg, msg + ", auto-generated (authorized dependency)");
		else if (dependencyProhibited(clientPkgName, supplierPkgName))
			addViolation(msg + ", reason=auto-detected (prohibited dependency)");
		else {
			if (isDebug) System.out.println(">>> fixViolation: prompting user");
			MessageBox box = new MessageBox(shell, SWT.YES | SWT.NO | SWT.CANCEL | SWT.APPLICATION_MODAL
					| SWT.ICON_QUESTION);
			box.setText("AccessViolations Module");
			box.setMessage("There is a(n) " + msg + ".\n\nDo you want " + "to create a Dependency from " + clientPkg.eClass().getName().toLowerCase() + ":\n\n   '"
					+ clientPkgName + "'\n\nto " + supplierPkg.eClass().getName().toLowerCase() + ":\n\n   '" + supplierPkgName + "'\n\n?");
			int reply = box.open();
			if (reply == SWT.YES)
				addDependency(clientPkg, supplierPkg, msg + ", user-request");
			else if (reply == SWT.CANCEL)
				throw new CancelException();
			else {
				illegalDependencies.add(clientPkgName + ">" + supplierPkgName);
				addViolation(msg + ", reason=disallowed by user");
			}
		}
	}

	private boolean dependencyAuthorized(String fromPkgName, String toPkgName) {
		if (!authorizedDependencies.isEmpty()) {
			for (int i = 0; i < authorizedDependencies.size(); i++) {
				String s = authorizedDependencies.get(i).toString();
				int pos = s.indexOf(">");
				if (pos > 0) {
					if (fromPkgName.startsWith(s.substring(0, pos)) && toPkgName.startsWith(s.substring(pos + 1)))
						return true;
				}
			}
		}
		if (globalLibraries.contains(toPkgName)) return true;
		return false;
	}

	private boolean dependencyProhibited(String fromPkgName, String toPkgName) {
		if (!prohibitedDependencies.isEmpty()) {
			for (int i = 0; i < prohibitedDependencies.size(); i++) {
				String s = prohibitedDependencies.get(i).toString();
				int pos = s.indexOf(">");
				if (pos > 0) {
					if (fromPkgName.startsWith(s.substring(0, pos)) && toPkgName.startsWith(s.substring(pos + 1)))
						return true;
				}
			}
		}
		return false;
	}

	private void addViolation(String inMsg) {
		numViolations++;
		if (!headerPrinted) {
			log("\n--- Processing " + curNSType + " '" + curNSName + "'");
			headerPrinted = true;
		}
		log("*** Illegal " + inMsg);
	}

	private String depInfo(Object elt, NamedElement from, NamedElement to) {
		String res = "";
		if (elt instanceof RSMRelation) {
			RSMRelation rel = (RSMRelation) elt;
			res = rel.getRel().eClass().getName().toLowerCase() + " from " + from.eClass().getName().toLowerCase()
					+ " '" + from.getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(from.getNamespace().getQualifiedName()) + "') to "
//					+ RSMModel.getNameWithoutModel(from.getNearestPackage().getQualifiedName()) + ") to "
					+ to.eClass().getName().toLowerCase() + " '" + to.getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(to.getNamespace().getQualifiedName()) + "')";
//					+ RSMModel.getNameWithoutModel(to.getNearestPackage().getQualifiedName()) + ")";
		} else if (elt instanceof Operation) {
			Operation op = (Operation) elt;
			Parameter p = (Parameter) to;
			String pName;
			if (p.getDirection().equals(ParameterDirectionKind.RETURN_LITERAL)) {
				pName = "return parameter";
			} else {
				pName = "parameter '" + p.getName() + "'";
			}
			res = "reference from " + pName + " in operation '" + op.getName() + "' of "
					+ from.eClass().getName().toLowerCase() + " '" + from.getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(from.getNamespace().getQualifiedName())
//					+ RSMModel.getNameWithoutModel(from.getNearestPackage().getQualifiedName())
					+ "') to parameter type '" + p.getType().getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(p.getType().getPackage().getQualifiedName()) + "')";
		} else if (elt instanceof Property) {
			Property att = (Property) elt;
			res = "reference from attribute '" + att.getName() + "' in " + from.eClass().getName().toLowerCase() + " '"
					+ from.getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(from.getNamespace().getQualifiedName())
//					+ RSMModel.getNameWithoutModel(from.getNearestPackage().getQualifiedName())
					+ "') to attribute type '" + att.getType().getName() + "' (in '"
					+ RSMModel.getNameWithoutModel(att.getType().getPackage().getQualifiedName()) + "')";
		}
		return res;
	}

	private void addDependency(Namespace clientPkg, Namespace supplierPkg, String msg) {
		String fromName = RSMModel.getNameWithoutModel(clientPkg.getQualifiedName());
		String toName = RSMModel.getNameWithoutModel(supplierPkg.getQualifiedName());
		if (!fromName.equalsIgnoreCase(toName)) {
			if (!headerPrinted) {
				log("\n--> Processing " + curNSType + " '" + curNSName + "'");
				headerPrinted = true;
			}
			log("+++ Resolving " + msg);
			createDependency(clientPkg, supplierPkg, msg);
			numFixedViolations = numFixedViolations + 1;
		}
	}

	private void createDependency(Namespace from, Namespace to, String reason) {
		String fromName = RSMModel.getNameWithoutModel(from.getQualifiedName());
		String toName = RSMModel.getNameWithoutModel(to.getQualifiedName());
		log("    Adding dependency from '" + from.getQualifiedName() + "' to '" + to.getQualifiedName() + "'");
		Dependency dep = from.createDependency(to);
		RSMModel.setDocumentation(dep, "*** Relation generated on " + RSMUtil.getDate(new Locale("en", "US"))
				+ ", based on " + reason + " (AccessViolations) ***");
		newDependencies.add(fromName + ">" + toName);
		hasChanged = true;
	}

	public int getTotViolations() {
		return totViolations;
	}

	public void setFixAccessViolations(boolean fixAccessViolations) {
		this.fixAccessViolations = fixAccessViolations;
	}

	/**
	 * @return
	 */
	public int getTotFixedViolations() {
		return totFixedViolations;
	}

	/**
	 * @return
	 */
	public int getNumPackages() {
		return numPackages;
	}
}
